@@ -1,20 +1,24 @@
-/*
-   Copyright (C) 2000-2001, 2003-2004 MySQL AB
-    All rights reserved. Use is subject to license terms.
+/* Copyright (c) 2000, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
-*/
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
 
 /* Calculate a checksum for a row */
 
@@ -28,7 +32,7 @@ ha_checksum mi_checksum(MI_INFO *info, const uchar *buf)
 
   for (i=info->s->base.fields ; i-- ; buf+=(rec++)->length)
   {
-    const uchar *pos;
+    uchar *pos;
     ulong length;
     switch (rec->type) {
     case FIELD_BLOB:
@@ -36,8 +40,7 @@ ha_checksum mi_checksum(MI_INFO *info, const uchar *buf)
       length=_mi_calc_blob_length(rec->length-
 					portable_sizeof_char_ptr,
 					buf);
-      memcpy((char*) &pos, buf+rec->length- portable_sizeof_char_ptr,
-	     sizeof(char*));
+      memcpy(&pos, buf+rec->length- portable_sizeof_char_ptr, sizeof(char*));
       break;
     }
     case FIELD_VARCHAR:
@@ -47,12 +50,12 @@ ha_checksum mi_checksum(MI_INFO *info, const uchar *buf)
         length= (ulong) *(uchar*) buf;
       else
         length= uint2korr(buf);
-      pos= buf+pack_length;
+      pos= (uchar*)buf+pack_length;
       break;
     }
     default:
       length=rec->length;
-      pos=buf;
+      pos= (uchar*)buf;
       break;
     }
     crc=my_checksum(crc, pos ? pos : (uchar*) "", length);
