@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -20,6 +27,9 @@
 
 #include "SignalData.hpp"
 
+#define JAM_FILE_ID 16
+
+
 class TcContinueB {
   /**
    * Sender(s)/Reciver(s)
@@ -43,11 +53,19 @@ private:
     ZWAIT_ABORT_ALL                        = 14,
     ZCHECK_SCAN_ACTIVE_FAILED_LQH          = 15,
     TRIGGER_PENDING                        = 17,
-    
-    DelayTCKEYCONF = 18,
-    ZNF_CHECK_TRANSACTIONS = 19,
-    ZSEND_FIRE_TRIG_REQ = 20
+    DelayTCKEYCONF                         = 18,
+    ZNF_CHECK_TRANSACTIONS                 = 19,
+    ZSEND_FIRE_TRIG_REQ                    = 20,
+    ZSTART_FRAG_SCANS                      = 21
+#ifdef ERROR_INSERT
+    ,ZDEBUG_DELAYED_ABORT                  = 22
+    ,ZDEBUG_DELAY_TCROLLBACKREP            = 23
+#endif
+    ,ZSCAN_FOR_READ_BACKUP                  = 24
   };
 };
 
+
+#undef JAM_FILE_ID
+
 #endif
