@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2016, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -140,6 +140,20 @@ Properties::Properties(const Property * anArray, int arrayLen){
   put(anArray, arrayLen);
 }
 
+Properties& Properties::operator=(const Properties & org){
+  if (this == &org) {
+    return *this;
+  }
+  delete impl;
+
+  propErrno = 0;
+  osErrno = 0;
+  impl = new PropertiesImpl(this, * org.impl);
+  parent = NULL;
+
+  return *this;
+}
+
 Properties::~Properties(){
   clear();
   delete impl;
@@ -525,7 +539,7 @@ void
 PropertiesImpl::setCaseInsensitiveNames(bool value){
   m_insensitive = value;
   if(value)
-    compare = strcasecmp;
+    compare = native_strcasecmp;
   else
     compare = strcmp;
 }
