@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2013, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2022, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -93,7 +93,9 @@ void Dbtup::execDBINFO_SCANREQ(Signal* signal)
   case Ndbinfo::POOLS_TABLEID:
   {
     jam();
-    Ndbinfo::pool_entry pools[] =
+    const DynArr256Pool::Info pmpInfo = c_page_map_pool.getInfo();
+    
+    const Ndbinfo::pool_entry pools[] =
     {
       { "Scan Lock",
         c_scanLockPool.getUsed(),
@@ -131,6 +133,26 @@ void Dbtup::execDBINFO_SCANREQ(Signal* signal)
         c_operation_pool.getEntrySize(),
         c_operation_pool.getUsedHi(),
         { CFG_DB_NO_LOCAL_OPS,CFG_DB_NO_OPS,0,0 }},
+      { "L2PMap pages",
+        pmpInfo.pg_count,
+        0,                  /* No real limit */
+        pmpInfo.pg_byte_sz,
+        /*
+          No HWM for this row as it would be a fixed fraction of "Data memory"
+          and therefore of limited interest.
+        */
+        0,
+        { 0, 0, 0}},
+      { "L2PMap nodes",
+        pmpInfo.inuse_nodes,
+        pmpInfo.pg_count * pmpInfo.nodes_per_page, /* Max within current pages */
+        pmpInfo.node_byte_sz,
+        /*
+          No HWM for this row as it would be a fixed fraction of "Data memory"
+          and therefore of limited interest.
+        */
+        0,
+        { 0, 0, 0 }},
       { "Data memory",
         m_pages_allocated,
         0, // Allocated from global resource group RG_DATAMEM
@@ -142,7 +164,9 @@ void Dbtup::execDBINFO_SCANREQ(Signal* signal)
 
     const size_t num_config_params =
       sizeof(pools[0].config_params) / sizeof(pools[0].config_params[0]);
+    const Uint32 numPools = NDB_ARRAY_SIZE(pools);
     Uint32 pool = cursor->data[0];
+    ndbrequire(pool < numPools);
     BlockNumber bn = blockToMain(number());
     while(pools[pool].poolname)
     {
@@ -211,8 +235,8 @@ void
 Dbtup::execDUMP_STATE_ORD(Signal* signal)
 {
   Uint32 type = signal->theData[0];
-  DumpStateOrd * const dumpState = (DumpStateOrd *)&signal->theData[0];
 
+  (void)type;
 #if 0
   if (type == 100) {
     RelTabMemReq * const req = (RelTabMemReq *)signal->getDataPtrSend();
@@ -277,6 +301,7 @@ Dbtup::execDUMP_STATE_ORD(Signal* signal)
 #endif
 #ifdef ERROR_INSERT
   if (type == DumpStateOrd::EnableUndoDelayDataWrite) {
+    DumpStateOrd * const dumpState = (DumpStateOrd *)&signal->theData[0];
     ndbout << "Dbtup:: delay write of datapages for table = " 
 	   << dumpState->args[1]<< endl;
     c_errorInsert4000TableId = dumpState->args[1];
@@ -389,6 +414,7 @@ Dbtup::execDUMP_STATE_ORD(Signal* signal)
   }
 #endif
 
+#ifdef ERROR_INSERT
   if (signal->theData[0] == DumpStateOrd::SchemaResourceSnapshot)
   {
     {
@@ -426,6 +452,7 @@ Dbtup::execDUMP_STATE_ORD(Signal* signal)
     RSS_AP_SNAPSHOT_CHECK2(c_storedProcPool, c_storedProcCountNonAPI);
     return;
   }
+#endif
 }//Dbtup::execDUMP_STATE_ORD()
 
 /* ---------------------------------------------------------------- */
@@ -500,18 +527,18 @@ operator<<(NdbOut& out, const Dbtup::Operationrec& op)
   // table
   out << " [fragmentPtr " << hex << op.fragmentPtr << "]";
   // type
-  out << " [op_type " << dec << op.op_struct.op_type << "]";
+  out << " [op_type " << dec << op.op_type << "]";
   out << " [delete_insert_flag " << dec;
-  out << op.op_struct.delete_insert_flag << "]";
+  out << op.op_struct.bit_field.delete_insert_flag << "]";
   // state
-  out << " [tuple_state " << dec << op.op_struct.tuple_state << "]";
-  out << " [trans_state " << dec << op.op_struct.trans_state << "]";
-  out << " [in_active_list " << dec << op.op_struct.in_active_list << "]";
+  out << " [tuple_state " << dec << op.tuple_state << "]";
+  out << " [trans_state " << dec << op.trans_state << "]";
+  out << " [in_active_list " << dec << op.op_struct.bit_field.in_active_list << "]";
   // links
   out << " [prevActiveOp " << hex << op.prevActiveOp << "]";
   out << " [nextActiveOp " << hex << op.nextActiveOp << "]";
   // tuples
-  out << " [tupVersion " << hex << op.tupVersion << "]";
+  out << " [tupVersion " << hex << op.op_struct.bit_field.tupVersion << "]";
   out << " [m_tuple_location " << op.m_tuple_location << "]";
   out << " [m_copy_tuple_location " << op.m_copy_tuple_location << "]";
   out << "]";
