@@ -127,6 +127,16 @@ struct PackedWordsContainer
   Uint32 packedWords[30];
 }; // 128 bytes
 
+#define LIGHT_LOAD_CONST 0
+#define MEDIUM_LOAD_CONST 1
+#define OVERLOAD_CONST 2
+enum OverloadStatus
+{
+  LIGHT_LOAD = LIGHT_LOAD_CONST,
+  MEDIUM_LOAD = MEDIUM_LOAD_CONST,
+  OVERLOAD = OVERLOAD_CONST
+};
+
 /**
   Description of NDB Software Architecture
   ----------------------------------------
@@ -492,6 +502,7 @@ public:
     EmulatedJamBuffer* jamBuffer;
     Uint32 * watchDogCounter;
     SectionSegmentPool::Cache * sectionPoolCache;
+    NDB_TICKS* pHighResTimer;
   };
   /* Setup state of a block object for executing in a particular thread. */
   void assignToThread(ThreadContext ctx);
@@ -591,6 +602,7 @@ public:
 
   void EXECUTE_DIRECT(ExecFunction f,
                       Signal *signal);
+
 protected:
   static Callback TheEmptyCallback;
   void TheNULLCallbackFunction(class Signal*, Uint32, Uint32);
@@ -598,10 +610,45 @@ protected:
   void execute(Signal* signal, Callback & c, Uint32 returnCode);
   
 
+  /**
+   * Various methods to get data from ndbd/ndbmtd such as time
+   * spent in sleep, sending and executing, number of signals
+   * in queue, and send buffer level.
+   *
+   * Also retrieving a thread name (this name must be pointing to a
+   * static pointer since it will be stored and kept for a long
+   * time. So the pointer cannot be changed.
+   *
+   * Finally also the ability to query for send thread information.
+   */
   void getSendBufferLevel(NodeId node, SB_LevelType &level);
   Uint32 getSignalsInJBB();
+  void setOverloadStatus(OverloadStatus new_status);
+  void setWakeupThread(Uint32 wakeup_instance);
+  void setNodeOverloadStatus(OverloadStatus new_status);
+  void setSendNodeOverloadStatus(OverloadStatus new_status);
+  void setNeighbourNode(NodeId node);
+  void getPerformanceTimers(Uint64 &micros_sleep,
+                            Uint64 &spin_time,
+                            Uint64 &buffer_full_micros_sleep,
+                            Uint64 &micros_send);
+  void getSendPerformanceTimers(Uint32 send_instance,
+                                Uint64 & exec_time,
+                                Uint64 & sleep_time,
+                                Uint64 & spin_time,
+                                Uint64 & user_time_os,
+                                Uint64 & kernel_time_os,
+                                Uint64 & elapsed_time_os);
+  Uint32 getSpintime();
+  Uint32 getNumSendThreads();
+  Uint32 getNumThreads();
+  const char * getThreadName();
+  const char * getThreadDescription();
 
-  NDB_TICKS getHighResTimer();
+  NDB_TICKS getHighResTimer() const 
+  {
+    return *m_pHighResTimer;
+  }
 
   /**********************************************************
    * Send signal - dialects
@@ -1007,7 +1054,7 @@ protected:
    * If the cause of the shutdown is known use extradata to add an 
    * errormessage describing the problem
    */
-  void progError(int line, int err_code, const char* extradata=NULL) const
+  void progError(int line, int err_code, const char* extradata=NULL, const char* check="") const
     ATTRIBUTE_NORETURN;
 private:
   void  signal_error(Uint32, Uint32, Uint32, const char*, int) const
@@ -1040,6 +1087,9 @@ private:
   /* For multithreaded ndb, the thread-specific watchdog counter. */
   Uint32 *m_watchDogCounter;
 
+  /* Read-only high res timer pointer */
+  const NDB_TICKS* m_pHighResTimer;
+
   SectionSegmentPool::Cache * m_sectionPoolCache;
   
   
@@ -1280,7 +1330,8 @@ public:
     BlockReference reference() const;
     void progError(int line,
                    int err_code,
-                   const char* extra = 0) ATTRIBUTE_NORETURN;
+                   const char* extra = 0,
+                   const char* check = "") ATTRIBUTE_NORETURN;
   };
   
   friend class MutexManager;
@@ -1384,6 +1435,32 @@ public:
   const char* debugOutTag(char* buf, int line);
 #endif
 
+  const char* getPartitionBalanceString(Uint32 fct)
+  {
+    switch (fct)
+    {
+      case NDB_PARTITION_BALANCE_SPECIFIC:
+        return "NDB_PARTITION_BALANCE_SPECIFIC";
+      case NDB_PARTITION_BALANCE_FOR_RA_BY_NODE:
+        return "NDB_PARTITION_BALANCE_FOR_RA_BY_NODE";
+      case NDB_PARTITION_BALANCE_FOR_RP_BY_NODE:
+        return "NDB_PARTITION_BALANCE_FOR_RP_BY_NODE";
+      case NDB_PARTITION_BALANCE_FOR_RP_BY_LDM:
+        return "NDB_PARTITION_BALANCE_FOR_RP_BY_LDM";
+      case NDB_PARTITION_BALANCE_FOR_RA_BY_LDM:
+        return "NDB_PARTITION_BALANCE_FOR_RA_BY_LDM";
+      case NDB_PARTITION_BALANCE_FOR_RA_BY_LDM_X_2:
+        return "NDB_PARTITION_BALANCE_FOR_RA_BY_LDM_X_2";
+      case NDB_PARTITION_BALANCE_FOR_RA_BY_LDM_X_3:
+        return "NDB_PARTITION_BALANCE_FOR_RA_BY_LDM_X_3";
+      case NDB_PARTITION_BALANCE_FOR_RA_BY_LDM_X_4:
+        return "NDB_PARTITION_BALANCE_FOR_RA_BY_LDM_X_4";
+      default:
+        ndbrequire(false);
+    }
+    return NULL;
+  }
+
   void ndbinfo_send_row(Signal* signal,
                         const DbinfoScanReq& req,
                         const Ndbinfo::Row& row,
