@@ -1,19 +1,26 @@
-/* Copyright (c) 2003-2005 MySQL AB
-
+/*
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA */
-
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 
 #include "API.hpp"
@@ -65,10 +72,9 @@ NdbApiSignal::NdbApiSignal(BlockReference ref)
 
 NdbApiSignal::NdbApiSignal(Ndb* ndb)
 {
-  BlockReference ref = ndb->theMyRef;
   theVerId_signalNumber   = 0;    // 4 bit ver id - 16 bit gsn
   theReceiversBlockNumber = 0;  // Only 16 bit blocknum  
-  theSendersBlockRef      = refToBlock(ref);
+  theSendersBlockRef      = ndb->theNdbBlockNumber;
   theLength               = 0;
   theSendersSignalId      = 0;
   theSignalId             = 0;
@@ -77,6 +83,8 @@ NdbApiSignal::NdbApiSignal(Ndb* ndb)
   m_fragmentInfo          = 0;
   for (int i = 0; i < 25; i++)
     theData[i] = 0x13579753;
+
+  assert(theSendersBlockRef != 0);
   
   setDataPtr(&theData[0]);
   theNextSignal = 0;
