@@ -1,19 +1,26 @@
-/* Copyright (c) 2003, 2005-2007 MySQL AB
-
+/*
+   Copyright (c) 2005, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA */
-
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <AttributeHeader.hpp>
@@ -593,7 +600,7 @@ NdbIndexStat::query_stat(const Range& range_f, Stat& stat_f)
   NdbIndexStatImpl::Bound& bound2 =
     *(NdbIndexStatImpl::Bound*)bound2_f.m_impl;
   NdbIndexStatImpl::Range range(bound1, bound2);
-#ifndef DBUG_OFF
+#ifndef NDEBUG
   const uint sz = 8000;
   char buf[sz];
   DBUG_PRINT("index_stat", ("lo: %s", bound1.m_bound.print(buf, sz)));
@@ -629,7 +636,7 @@ NdbIndexStat::get_rir(const Stat& stat_f, double* rir)
     x = 1.0;
   require(rir != 0);
   *rir = x;
-#ifndef DBUG_OFF
+#ifndef NDEBUG
   char buf[100];
   sprintf(buf, "%.2f", *rir);
 #endif
@@ -648,7 +655,7 @@ NdbIndexStat::get_rpk(const Stat& stat_f, Uint32 k, double* rpk)
     x = 1.0;
   require(rpk != 0);
   *rpk = x;
-#ifndef DBUG_OFF
+#ifndef NDEBUG
   char buf[100];
   sprintf(buf, "%.2f", *rpk);
 #endif
@@ -706,6 +713,15 @@ NdbIndexStat::create_listener(Ndb* ndb)
   DBUG_RETURN(0);
 }
 
+bool
+NdbIndexStat::has_listener() const
+{
+  DBUG_ENTER("NdbIndexStat::has_listener");
+  if (m_impl.m_eventOp != 0)
+    DBUG_RETURN(true);
+  DBUG_RETURN(false);
+}
+
 int
 NdbIndexStat::execute_listener(Ndb* ndb)
 {
@@ -739,8 +755,7 @@ int
 NdbIndexStat::drop_listener(Ndb* ndb)
 {
   DBUG_ENTER("NdbIndexStat::drop_listener");
-  if (m_impl.drop_listener(ndb) == -1)
-    DBUG_RETURN(-1);
+  (void)m_impl.drop_listener(ndb);
   DBUG_RETURN(0);
 }
 
