@@ -1,31 +1,29 @@
 /*
-   Copyright (c) 2003-2008 MySQL AB, 2008, 2009 Sun Microsystems, Inc.
-   Use is subject to license terms.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
 */
 
-#include <ndb_global.h>
-#include <NdbOperation.hpp>
-#include "NdbApiSignal.hpp"
-#include <NdbTransaction.hpp>
-#include <Ndb.hpp>
-#include <NdbRecAttr.hpp>
-#include "NdbUtil.hpp"
+#include "API.hpp"
 #include "NdbOut.hpp"
-#include "NdbImpl.hpp"
-#include <NdbIndexScanOperation.hpp>
 #include <NdbBlob.hpp>
 
 #include <Interpreter.hpp>
@@ -387,7 +385,10 @@ NdbOperation::getValue_impl(const NdbColumnImpl* tAttrInfo, char* aValue)
   NdbRecAttr* tRecAttr;
   if ((tAttrInfo != NULL) &&
       (theStatus != Init)){
-    m_no_disk_flag &= (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+    if (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK)
+    {
+      m_flags &= ~Uint8(OF_NO_DISK);
+    }
     if (theStatus != GetValue) {
       if (theStatus == UseNdbRecord)
         /* This path for extra GetValues for NdbRecord */
@@ -446,7 +447,10 @@ NdbOperation::getValue_NdbRecord(const NdbColumnImpl* tAttrInfo, char* aValue)
 {
   NdbRecAttr* tRecAttr;
 
-  m_no_disk_flag &= (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+  if (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK)
+  {
+    m_flags &= ~Uint8(OF_NO_DISK);
+  }
 
   /*
     For getValue with NdbRecord operations, we just allocate the NdbRecAttr,
@@ -559,7 +563,10 @@ NdbOperation::setValue( const NdbColumnImpl* tAttrInfo,
   
   // Insert Attribute Id into ATTRINFO part. 
   tAttrId = tAttrInfo->m_attrId;
-  m_no_disk_flag &= (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+  if (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK)
+  {
+    m_flags &= ~Uint8(OF_NO_DISK);
+  }
   const char *aValue = aValuePassed; 
   if (aValue == NULL) {
     if (tAttrInfo->m_nullable) {
@@ -578,24 +585,20 @@ NdbOperation::setValue( const NdbColumnImpl* tAttrInfo,
     }//if
   }//if
   
-  Uint32 sigLen = 0;  // Length of significant part
-  Uint32 sendLen = 0; // Length of data to send.  Normally sigLen == sendLen
-
-  if (! tAttrInfo->get_var_length_bug39645(aValue, sigLen, sendLen)) {
+  Uint32 len;
+  if (! tAttrInfo->get_var_length(aValue, len)) {
     setErrorCodeAbort(4209);
     DBUG_RETURN(-1);
   }
-  
-  const Uint32 sizeInBytes = sendLen;
+
+  const Uint32 sizeInBytes = len;
   const Uint32 bitsInLastWord = 8 * (sizeInBytes & 3) ;
   
   const int attributeSize = sizeInBytes;
   const int slack = sizeInBytes & 3;
   
-  if (((UintPtr)aValue & 3) != 0 || // Alignment 
-      (slack != 0) ||               // Slack bits
-      (sigLen != sendLen)){         // Bug39645
-    memcpy(&tempData[0], aValue, sigLen);
+  if (((UintPtr)aValue & 3) != 0 || (slack != 0)){
+    memcpy(&tempData[0], aValue, attributeSize);
     aValue = (char*)&tempData[0];
     if(slack != 0) {
       char * tmp = (char*)&tempData[0];
@@ -736,6 +739,7 @@ NdbOperation::getBlobHandle(NdbTransaction* aCon, const NdbColumnImpl* tAttrInfo
     tLastBlob->theNext = tBlob;
   tBlob->theNext = NULL;
   theNdbCon->theBlobFlag = true;
+  theNdbCon->m_userDefinedBlobOps = true;
   return tBlob;
 }
 
@@ -813,6 +817,7 @@ NdbOperation::linkInBlobHandle(NdbTransaction *aCon,
   lastPtr= bh;
   bh->theNext= NULL;
   theNdbCon->theBlobFlag= true;
+  theNdbCon->m_userDefinedBlobOps = true;
 
   return bh;
 }
@@ -825,7 +830,8 @@ NdbOperation::linkInBlobHandle(NdbTransaction *aCon,
  * For read request, store the pointers to blob handles in the row.
  */
 int
-NdbOperation::getBlobHandlesNdbRecord(NdbTransaction* aCon)
+NdbOperation::getBlobHandlesNdbRecord(NdbTransaction* aCon, 
+                                      const Uint32 * m_read_mask)
 {
   NdbBlob *lastBlob= NULL;
 
@@ -868,7 +874,8 @@ NdbOperation::getBlobHandlesNdbRecord(NdbTransaction* aCon)
 */
 int
 NdbOperation::getBlobHandlesNdbRecordDelete(NdbTransaction* aCon,
-                                            bool checkReadSet)
+                                            bool checkReadSet,
+                                            const Uint32 * m_read_mask)
 {
   NdbBlob *lastBlob= NULL;
 
@@ -953,7 +960,7 @@ NdbOperation::insertATTRINFO( Uint32 aData )
     tAI_LenInCurrAI = 3;
     tSignal = tNdb->getSignal();
     if (tSignal != NULL) {
-      tSignal->setSignal(m_attrInfoGSN);
+      tSignal->setSignal(m_attrInfoGSN, refToBlock(theNdbCon->m_tcRef));
       tAttrPtr = &tSignal->getDataPtrSend()[3];
       if (tFirstAttrinfo == NULL) {
         tSignal->next(NULL);
@@ -1010,7 +1017,7 @@ NdbOperation::insertATTRINFOloop(register const Uint32* aDataPtr,
       tAI_LenInCurrAI = 3;
       tSignal = tNdb->getSignal();
       if (tSignal != NULL) {
-        tSignal->setSignal(m_attrInfoGSN);
+        tSignal->setSignal(m_attrInfoGSN, refToBlock(theNdbCon->m_tcRef));
         tAttrPtr = &tSignal->getDataPtrSend()[3];
         if (tFirstAttrinfo == NULL) {
           tSignal->next(NULL);
@@ -1076,6 +1083,42 @@ NdbOperation::setAbortOption(AbortOption ao)
 }
 
 
+int
+NdbOperation::prepareGetLockHandleNdbRecord()
+{
+  /* This method is used to perform the correct actions
+   * when the OO_LOCKHANDLE flag is set on an NdbRecord
+   * operation.
+   */
+  assert(theLockHandle == NULL);
+  theLockHandle = theNdbCon->getLockHandle();
+  if (!theLockHandle)
+  {
+    return 4000; /* Memory allocation issue */
+  }
+
+  assert(! theLockHandle->isLockRefValid());
+
+  assert(m_attribute_record);
+  theLockHandle->m_table = m_attribute_record->table;
+  assert(theLockHandle->m_table);
+  
+  NdbRecAttr* ra = 
+    getValue_NdbRecord(&NdbColumnImpl::getImpl(*NdbDictionary::Column::LOCK_REF),
+                       (char*) &theLockHandle->m_lockRef);
+  
+  if (!ra)
+  {
+    /* Assume error code set */
+    assert(theError.code);
+    return theError.code;
+  }
+
+  theLockHandle->m_state = NdbLockHandle::PREPARED;
+
+  return 0;
+}
+
 /*
  * handleOperationOptions
  * static member for setting operation options
@@ -1269,6 +1312,17 @@ NdbOperation::handleOperationOptions (const OperationType type,
        * takeover operation 
        */
     }
+    /* Only allowed for pk ops on user defined partitioned tables
+     * or when defining an unlock operation
+     */
+    if (unlikely( ! (((op->m_attribute_record->flags & 
+                       NdbRecord::RecHasUserDefinedPartitioning) &&
+                      (op->m_key_record->table->m_index == NULL)) ||
+                     (type == UnlockRequest))))
+    {
+      /* Explicit partitioning info not allowed for table and operation*/
+      return 4546;
+    }
     op->theDistributionKey=opts->partitionId;
     op->theDistrKeyIndicator_= 1;       
   }
@@ -1310,7 +1364,7 @@ NdbOperation::handleOperationOptions (const OperationType type,
   {
     /* Any operation can have an ANYVALUE set */
     op->m_any_value = opts->anyValue;
-    op->m_use_any_value = 1;
+    op->m_flags |= OF_USE_ANY_VALUE;
   }
 
   if (opts->optionsPresent & OperationOptions::OO_CUSTOMDATA)
@@ -1319,5 +1373,55 @@ NdbOperation::handleOperationOptions (const OperationType type,
     op->m_customData = opts->customData;
   }
 
+  if (opts->optionsPresent & OperationOptions::OO_LOCKHANDLE)
+  {
+    if (unlikely(op->theNdb->getMinDbNodeVersion() <
+                 NDBD_UNLOCK_OP_SUPPORTED))
+    {
+      /* Function not implemented yet */
+      return 4003;
+    }
+
+    /* Check that this is a pk read with a lock 
+     * No need to worry about Blob lock upgrade issues as
+     * Blobs have not been handled at this stage
+     */
+    if (((type != ReadRequest) &&
+         (type != ReadExclusive)) ||
+        (op->m_key_record &&
+         (op->m_key_record->flags & NdbRecord::RecIsIndex)) ||
+        ((op->theLockMode != LM_Read) &&
+         (op->theLockMode != LM_Exclusive)))
+    {
+      return 4549; /* getLockHandle only supported for primary key read with a lock */
+    }
+
+    int prepareRc = op->prepareGetLockHandleNdbRecord();      
+    if (prepareRc != 0)
+    {
+      return prepareRc;
+    }    
+  }
+
+  if (opts->optionsPresent & OperationOptions::OO_QUEUABLE)
+  {
+    op->m_flags |= OF_QUEUEABLE;
+  }
+
+  if (opts->optionsPresent & OperationOptions::OO_NOT_QUEUABLE)
+  {
+    op->m_flags &= ~Uint8(OF_QUEUEABLE);
+  }
+
+  if (opts->optionsPresent & OperationOptions::OO_DEFERRED_CONSTAINTS)
+  {
+    op->m_flags |= OF_DEFERRED_CONSTRAINTS;
+  }
+
+  if (opts->optionsPresent & OperationOptions::OO_DISABLE_FK)
+  {
+    op->m_flags |= OF_DISABLE_FK;
+  }
+
   return 0;
 }
