@@ -1,19 +1,26 @@
-/* Copyright (c) 2003-2007 MySQL AB
-
+/*
+   Copyright (c) 2004, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA */
-
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 
@@ -31,6 +38,7 @@
 #include <mgmapi_internal.h>
 #include "NdbImpl.hpp"
 #include "NdbDictionaryImpl.hpp"
+#include "ProcessInfo.hpp"
 
 #include <NdbMutex.h>
 #ifdef VM_TRACE
@@ -65,6 +73,16 @@ Ndb_cluster_connection::Ndb_cluster_connection(const char *connect_string,
 {
 }
 
+
+Ndb_cluster_connection::Ndb_cluster_connection(const char *connect_string,
+                                               Ndb_cluster_connection *
+                                               main_connection,
+                                               int force_api_nodeid)
+  : m_impl(* new Ndb_cluster_connection_impl(connect_string,
+                                             main_connection, force_api_nodeid))
+{
+}
+
 Ndb_cluster_connection::Ndb_cluster_connection
 (Ndb_cluster_connection_impl& impl) : m_impl(impl)
 {
@@ -91,6 +109,51 @@ const char *Ndb_cluster_connection::get_connected_host() const
   return 0;
 }
 
+int
+Ndb_cluster_connection::unset_recv_thread_cpu(Uint32 recv_thread_id)
+{
+  if (m_impl.m_transporter_facade)
+  {
+    return m_impl.m_transporter_facade->unset_recv_thread_cpu(recv_thread_id);
+  }
+  return -1;
+}
+
+int
+Ndb_cluster_connection::set_recv_thread_cpu(Uint16 *cpuid_array,
+                                            Uint32 array_len,
+                                            Uint32 recv_thread_id)
+{
+  if (m_impl.m_transporter_facade)
+  {
+    return m_impl.m_transporter_facade->set_recv_thread_cpu(cpuid_array,
+                                                            array_len,
+                                                            recv_thread_id);
+  }
+  return -1;
+}
+
+int
+Ndb_cluster_connection::set_recv_thread_activation_threshold(Uint32 threshold)
+{
+  TransporterFacade *fac = m_impl.m_transporter_facade;
+  if (fac)
+  {
+    return fac->set_recv_thread_activation_threshold(threshold);
+  }
+  return -1;
+}
+
+int
+Ndb_cluster_connection::get_recv_thread_activation_threshold() const
+{
+  if (m_impl.m_transporter_facade)
+  {
+    return m_impl.m_transporter_facade->get_recv_thread_activation_threshold();
+  }
+  return -1;
+}
+
 const char *Ndb_cluster_connection::get_connectstring(char *buf,
 						      int buf_sz) const
 {
@@ -157,17 +220,29 @@ Ndb_cluster_connection_impl::init_get_next_node
 Uint32
 Ndb_cluster_connection_impl::get_next_node(Ndb_cluster_connection_node_iter &iter)
 {
+  /**
+   * Note that iter may be out dated by changes to m_nodes_proximity.
+   * m_nodes_proximity can be changed by application calling
+   * set_data_node_neighbour() which in turn calls adjust_node_proximity that
+   * can rearrange the nodes.  This can even happen concurrently by another
+   * thread.
+   *
+   * It is assumed that each fields in the Node struct will at least be updated
+   * atomically.  And the fact that sometimes the next node selected may be the
+   * wrong one is ignored and taken as an glitch choosing a possible non
+   * optimal node once after call to set_data_node_neighbour().
+   */
   Uint32 cur_pos= iter.cur_pos;
   if (cur_pos >= no_db_nodes())
     return 0;
 
-  Ndb_cluster_connection_impl::Node *nodes= m_all_nodes.getBase();
+  Ndb_cluster_connection_impl::Node *nodes= m_nodes_proximity.getBase();
   Ndb_cluster_connection_impl::Node &node=  nodes[cur_pos];
 
   if (iter.scan_state != (Uint8)~0)
   {
     assert(iter.scan_state < no_db_nodes());
-    if (nodes[iter.scan_state].group == node.group)
+    if (nodes[iter.scan_state].adjusted_group == node.adjusted_group)
       iter.scan_state= ~0;
     else
       return nodes[iter.scan_state++].id;
@@ -177,9 +252,9 @@ Ndb_cluster_connection_impl::get_next_node(Ndb_cluster_connection_node_iter &ite
 
   cur_pos++;
   Uint32 init_pos= iter.init_pos;
-  if (cur_pos == node.next_group)
+  if (cur_pos == node.next_group_idx)
   {
-    cur_pos= nodes[init_pos].this_group;
+    cur_pos= nodes[init_pos].this_group_idx;
   }
 
   //  fprintf(stderr,"[cur_pos %d]",cur_pos);
@@ -187,8 +262,8 @@ Ndb_cluster_connection_impl::get_next_node(Ndb_cluster_connection_node_iter &ite
     iter.cur_pos= cur_pos;
   else
   {
-    iter.cur_pos= node.next_group;
-    iter.init_pos= node.next_group;
+    iter.cur_pos= node.next_group_idx;
+    iter.init_pos= node.next_group_idx;
   }
   return node.id;
 }
@@ -204,13 +279,13 @@ Ndb_cluster_connection_impl::get_next_alive_node(Ndb_cluster_connection_node_ite
 
   while ((id = get_next_node(iter)))
   {
-    tp->lock_mutex();
+    tp->lock_poll_mutex();
     if (tp->get_node_alive(id) != 0)
     {
-      tp->unlock_mutex();
+      tp->unlock_poll_mutex();
       return id;
     }
-    tp->unlock_mutex();
+    tp->unlock_poll_mutex();
   }
   return 0;
 }
@@ -218,7 +293,9 @@ Ndb_cluster_connection_impl::get_next_alive_node(Ndb_cluster_connection_node_ite
 unsigned
 Ndb_cluster_connection::no_db_nodes()
 {
-  return m_impl.m_all_nodes.size();
+  assert(m_impl.m_db_nodes.count() ==
+           m_impl.m_nodes_proximity.size());
+  return m_impl.m_nodes_proximity.size();
 }
 
 unsigned
@@ -234,18 +311,20 @@ Ndb_cluster_connection::max_nodegroup()
   if (tp == 0 || tp->ownId() == 0)
     return 0;
 
-  Bitmask<MAX_NDB_NODES> ng;
-  tp->lock_mutex();
-  for(unsigned i= 0; i < no_db_nodes(); i++)
+  NdbNodeBitmask ng;
+  tp->lock_poll_mutex();
+  for (Uint32 node_id = m_impl.m_db_nodes.find_first();
+       node_id != NdbNodeBitmask::NotFound;
+       node_id = m_impl.m_db_nodes.find_next(node_id + 1))
   {
     //************************************************
     // If any node is answering, ndb is answering
     //************************************************
-    trp_node n = tp->theClusterMgr->getNodeInfo(m_impl.m_all_nodes[i].id);
+    trp_node n = tp->theClusterMgr->getNodeInfo(node_id);
     if (n.is_confirmed() && n.m_state.nodeGroup <= MAX_NDB_NODES)
       ng.set(n.m_state.nodeGroup);
   }
-  tp->unlock_mutex();
+  tp->unlock_poll_mutex();
 
   if (ng.isclear())
     return 0;
@@ -267,17 +346,20 @@ int Ndb_cluster_connection::get_no_ready()
     return -1;
 
   unsigned int foundAliveNode = 0;
-  tp->lock_mutex();
-  for(unsigned i= 0; i < no_db_nodes(); i++)
+  tp->lock_poll_mutex();
+  for (Uint32 node_id = m_impl.m_db_nodes.find_first();
+       node_id != NdbNodeBitmask::NotFound;
+       node_id = m_impl.m_db_nodes.find_next(node_id + 1))
   {
     //************************************************
     // If any node is answering, ndb is answering
     //************************************************
-    if (tp->get_node_alive(m_impl.m_all_nodes[i].id) != 0) {
+    if (tp->get_node_alive(node_id) != 0)
+    {
       foundAliveNode++;
     }
   }
-  tp->unlock_mutex();
+  tp->unlock_poll_mutex();
 
   return foundAliveNode;
 }
@@ -300,13 +382,13 @@ Ndb_cluster_connection::wait_until_ready(int timeout,
   int milliCounter = 0;
   int noChecksSinceFirstAliveFound = 0;
   do {
-    unsigned int foundAliveNode = get_no_ready();
+    const Uint32 unconnected_nodes = m_impl.get_unconnected_nodes();
 
-    if (foundAliveNode == no_db_nodes())
+    if (unconnected_nodes == 0)
     {
       DBUG_RETURN(0);
     }
-    else if (foundAliveNode > 0)
+    else if (unconnected_nodes < no_db_nodes())
     {
       noChecksSinceFirstAliveFound++;
       // 100 ms delay -> 10*
@@ -362,8 +444,12 @@ Ndb_cluster_connection_impl(const char * connect_string,
     m_first_ndb_object(0),
     m_latest_error_msg(),
     m_latest_error(0),
-    m_max_trans_id(0),
-    m_multi_wait_group(0)
+    m_data_node_neighbour(0),
+    m_multi_wait_group(0),
+    m_uri_scheme(NULL),
+    m_uri_host(NULL),
+    m_uri_path(NULL),
+    m_uri_port(0)
 {
   DBUG_ENTER("Ndb_cluster_connection");
   DBUG_PRINT("enter",("Ndb_cluster_connection this=0x%lx", (long) this));
@@ -391,6 +477,8 @@ Ndb_cluster_connection_impl(const char * connect_string,
 
   m_event_add_drop_mutex= NdbMutex_Create();
   m_new_delete_ndb_mutex = NdbMutex_Create();
+  m_new_delete_ndb_cond = NdbCondition_Create();
+  m_nodes_proximity_mutex = NdbMutex_Create();
 
   m_connect_thread= 0;
   m_connect_callback= 0;
@@ -421,9 +509,10 @@ Ndb_cluster_connection_impl(const char * connect_string,
     m_transporter_facade=
       new TransporterFacade(m_main_connection->m_impl.m_globalDictCache);
 
-    // The secondary connection can't use same nodeid, clear the nodeid
-    // in ConfigRetriver to avoid asking for the same nodeid again
-    m_config_retriever->setNodeId(0);
+    // The secondary connection can't use same nodeid, but it's ok
+    // to specify one to use. Use the force_api_nodeid given(although
+    // it will normally be 0 in order to allocate dynamic nodeid)
+    m_config_retriever->setNodeId(force_api_nodeid);
 
   }
 
@@ -434,20 +523,23 @@ Ndb_cluster_connection_impl::~Ndb_cluster_connection_impl()
 {
   DBUG_ENTER("~Ndb_cluster_connection");
 
-  if (m_first_ndb_object != 0)
+  // Wait until all Ndb instances belonging to this Ndb_cluster_connection
+  // have been released(they have references to the TransporterFacade)
+  NdbMutex_Lock(m_new_delete_ndb_mutex);
+  if (m_first_ndb_object)
   {
-    g_eventLogger->warning("Deleting Ndb_cluster_connection with Ndb-object"
-                           " not deleted");
-    Ndb * p = m_first_ndb_object;
-    printf("this: %p Ndb-object(s): ", (Ndb_cluster_connection*)this);
-    while (p)
+    g_eventLogger->warning("Waiting for Ndb instances belonging to "
+                           "Ndb_cluster_connection %p to be deleted...",
+                           this);
+
+    while(m_first_ndb_object)
     {
-      printf("%p ", p);
-      p = p->theImpl->m_next_ndb_object;
+      NdbCondition_WaitTimeout(m_new_delete_ndb_cond,
+                               m_new_delete_ndb_mutex,
+                               1000);
     }
-    printf("\n");
-    fflush(stdout);
   }
+  NdbMutex_Unlock(m_new_delete_ndb_mutex);
 
   if (m_transporter_facade != 0)
   {
@@ -489,6 +581,12 @@ Ndb_cluster_connection_impl::~Ndb_cluster_connection_impl()
   }
   NdbMutex_Unlock(g_ndb_connection_mutex);
 
+  if (m_nodes_proximity_mutex != NULL)
+  {
+    NdbMutex_Destroy(m_nodes_proximity_mutex);
+    m_nodes_proximity_mutex = NULL;
+  }
+
   if (m_event_add_drop_mutex)
     NdbMutex_Destroy(m_event_add_drop_mutex);
   m_event_add_drop_mutex = 0;
@@ -496,22 +594,30 @@ Ndb_cluster_connection_impl::~Ndb_cluster_connection_impl()
   if (m_new_delete_ndb_mutex)
     NdbMutex_Destroy(m_new_delete_ndb_mutex);
   m_new_delete_ndb_mutex = 0;
+
+  if (m_new_delete_ndb_cond)
+    NdbCondition_Destroy(m_new_delete_ndb_cond);
+  m_new_delete_ndb_cond = 0;
   
   if(m_multi_wait_group)
     delete m_multi_wait_group;
   m_multi_wait_group = 0;
 
+  m_uri_scheme.clear();
+  m_uri_path.clear();
+  m_uri_host.clear();
+
   DBUG_VOID_RETURN;
 }
 
 void
-Ndb_cluster_connection::lock_ndb_objects()
+Ndb_cluster_connection::lock_ndb_objects() const
 {
   NdbMutex_Lock(m_impl.m_new_delete_ndb_mutex);
 }
 
 void
-Ndb_cluster_connection::unlock_ndb_objects()
+Ndb_cluster_connection::unlock_ndb_objects() const
 {
   NdbMutex_Unlock(m_impl.m_new_delete_ndb_mutex);
 }
@@ -537,7 +643,9 @@ Ndb_cluster_connection_impl::link_ndb_object(Ndb* p)
   p->theImpl->m_next_ndb_object = m_first_ndb_object;
   m_first_ndb_object = p;
   
-  p->theFirstTransId += m_max_trans_id;
+  // Wake up anyone waiting for changes to the Ndb instance list
+  NdbCondition_Broadcast(m_new_delete_ndb_cond);
+
   unlock_ndb_objects();
 }
 
@@ -566,12 +674,6 @@ Ndb_cluster_connection_impl::unlink_ndb_object(Ndb* p)
   p->theImpl->m_prev_ndb_object = 0;
   p->theImpl->m_next_ndb_object = 0;
 
-  Uint32 transId = (Uint32)p->theFirstTransId;
-  if (transId > m_max_trans_id)
-  {
-    m_max_trans_id = transId;
-  }
-
   /* This Ndb is leaving for a better place,
    * record its contribution to global warming
    * for posterity
@@ -581,9 +683,233 @@ Ndb_cluster_connection_impl::unlink_ndb_object(Ndb* p)
     globalApiStatsBaseline[i] += p->theImpl->clientStats[i];
   }
 
+  // Wake up anyone waiting for changes to the Ndb instance list
+  NdbCondition_Broadcast(m_new_delete_ndb_cond);
+
   unlock_ndb_objects();  
 }
 
+Uint32
+Ndb_cluster_connection_impl::get_next_transid(Uint32 reference) const
+{    
+  Uint32 next = 0;
+  reference = m_transporter_facade->mapRefToIdx(reference);
+  
+  /* Access map with lock to avoid resize issues */
+  lock_ndb_objects();
+  {
+    if (reference < m_next_transids.size())
+    {
+      next = m_next_transids[reference];
+    }
+  }
+  unlock_ndb_objects();
+  
+  return next;
+}
+
+void 
+Ndb_cluster_connection_impl::set_next_transid(Uint32 reference, Uint32 value)
+{
+  reference = m_transporter_facade->mapRefToIdx(reference);
+  
+  /* Access map with lock to avoid resize issues */
+  lock_ndb_objects();
+  {
+    Uint32 zero = 0;
+    m_next_transids.set(value, reference, zero);
+  }
+  unlock_ndb_objects();
+}
+
+/**
+ * adjust_node_proximity
+ *
+ * A negative adjustment means nearer.
+ *
+ * May rearrange m_nodes_proximity and change links and group value.
+ * The vector m_nodes_proximity itself, including size(), is not modified
+ * only the elements within.
+ *
+ * m_nodes_proximity_mutex must be locked and m_nodes_proximity_seqlock
+ * must be locked for write.
+ */
+void
+Ndb_cluster_connection_impl::adjust_node_proximity(Uint32 node_id, Int32 adjustment)
+{
+  assert(m_db_nodes.get(node_id)); // Ensured in set_data_node_neighbour
+
+  if (adjustment == 0)
+    return; // No change
+
+  Uint32 old_idx;
+  for (old_idx = 0; old_idx < m_nodes_proximity.size(); old_idx++)
+  {
+    if (m_nodes_proximity[old_idx].id == node_id)
+      break;
+  }
+  require(old_idx < m_nodes_proximity.size());
+
+  const Int32 old_group = m_nodes_proximity[old_idx].adjusted_group;
+  const Int32 new_group = old_group + adjustment;
+  Node node = m_nodes_proximity[old_idx];
+  node.adjusted_group = new_group;
+
+  Uint32 new_idx;
+  if (adjustment < 0)
+  {
+    /**
+     * Node is moved to be new last in its new group.
+     */
+    for (new_idx = 0; new_idx < old_idx; new_idx ++)
+    {
+      if (m_nodes_proximity[new_idx].adjusted_group > new_group)
+        break;
+    }
+    /**
+     * Move nodes between new_idx (inclusive) and old_idx (exclusive) up,
+     * making room for node in its new group.
+     */
+    for (Uint32 idx = old_idx;
+         idx > new_idx;
+         idx--)
+    {
+      m_nodes_proximity[idx] = m_nodes_proximity[idx-1];
+      m_nodes_proximity[idx].this_group_idx++;
+      if (m_nodes_proximity[idx].next_group_idx > 0 &&
+          m_nodes_proximity[idx].next_group_idx <= old_idx)
+      {
+        m_nodes_proximity[idx].next_group_idx++;
+        if (m_nodes_proximity[idx].next_group_idx == m_nodes_proximity.size())
+        {
+          m_nodes_proximity[idx].next_group_idx = 0;
+        }
+      }
+    }
+    /**
+     * For elements after old place with same group, this_group_idx needs
+     * increase.
+     */
+    for (Uint32 idx = old_idx + 1;
+         idx < m_nodes_proximity.size() &&
+           m_nodes_proximity[idx].adjusted_group == old_group;
+         idx++)
+    {
+      m_nodes_proximity[idx].this_group_idx++;
+    }
+    /**
+     * Update this_group_idx and next_group_idx for node.
+     */
+    if (new_idx == 0)
+    {
+      node.this_group_idx = 0;
+    }
+    else if (m_nodes_proximity[new_idx - 1].adjusted_group == new_group)
+    {
+      node.this_group_idx = m_nodes_proximity[new_idx - 1].this_group_idx;
+    }
+    else
+    {
+      node.this_group_idx = new_idx;
+    }
+    Uint32 next_group_idx = new_idx + 1;
+    if (next_group_idx < m_nodes_proximity.size())
+    {
+      node.next_group_idx = next_group_idx;
+    }
+    else
+    {
+      node.next_group_idx = 0;
+    }
+  }
+  else
+  {
+    /**
+     * Node is moved to be first in its new group.
+     */
+    for (new_idx = old_idx; new_idx + 1 < m_nodes_proximity.size(); new_idx++)
+    {
+      if (m_nodes_proximity[new_idx + 1].adjusted_group >= new_group)
+        break;
+    }
+    /**
+     * Move nodes between old_idx (exclusive) and new_idx (inclusive) down,
+     * making room for node in its new group.
+     */
+    for (Uint32 idx = old_idx;
+         idx < new_idx;
+         idx++)
+    {
+      m_nodes_proximity[idx] = m_nodes_proximity[idx + 1];
+      if (m_nodes_proximity[idx].this_group_idx > old_idx)
+      {
+        m_nodes_proximity[idx].this_group_idx--;
+      }
+      if (m_nodes_proximity[idx].next_group_idx > 0 &&
+          m_nodes_proximity[idx].next_group_idx < new_idx)
+      {
+        m_nodes_proximity[idx].next_group_idx--;
+      }
+      else
+      {
+        m_nodes_proximity[idx].next_group_idx = new_idx;
+      }
+    }
+    /**
+     * Update this_group_idx and next_group_idx for node.
+     */
+    if (old_idx < new_idx)
+    {
+      node.this_group_idx = new_idx;
+    }
+    if (new_idx + 1 == m_nodes_proximity.size())
+    {
+      node.next_group_idx = 0;
+    }
+    else if (m_nodes_proximity[new_idx + 1].adjusted_group == new_group)
+    {
+      node.next_group_idx = m_nodes_proximity[new_idx + 1].next_group_idx;
+    }
+    else
+    {
+      node.next_group_idx = new_idx + 1;
+    }
+  }
+  m_nodes_proximity[new_idx] = node;
+
+  /**
+   * Clear hint count in new group since the node adjusted will not have a
+   * hint count in sync with its new group.
+   */
+  for (Uint32 idx = node.this_group_idx;
+       idx <= new_idx;
+       idx++)
+  {
+    m_nodes_proximity[idx].hint_count = 0;
+  }
+}
+
+void
+Ndb_cluster_connection_impl::set_data_node_neighbour(Uint32 node)
+{
+  Uint32 const old_node = m_data_node_neighbour;
+  if (old_node == node)
+    return; // No change
+
+  NdbMutex_Lock(m_nodes_proximity_mutex);
+  if (old_node != 0 && m_db_nodes.get(old_node))
+  {
+    adjust_node_proximity(old_node,
+                          -DATA_NODE_NEIGHBOUR_PROXIMITY_ADJUSTMENT);
+  }
+  if (node != 0 && m_db_nodes.get(node))
+  {
+    adjust_node_proximity(node, DATA_NODE_NEIGHBOUR_PROXIMITY_ADJUSTMENT);
+  }
+  m_data_node_neighbour = node;
+  NdbMutex_Unlock(m_nodes_proximity_mutex);
+}
+
 void
 Ndb_cluster_connection_impl::set_name(const char *name)
 {
@@ -591,6 +917,41 @@ Ndb_cluster_connection_impl::set_name(const char *name)
   ndb_mgm_set_name(h, name);
 }
 
+int
+Ndb_cluster_connection_impl::set_service_uri(const char * scheme,
+                                             const char * host,
+                                             int port, const char * path)
+{
+  if(! ProcessInfo::isValidUri(scheme, path))
+  {
+    return 1;
+  }
+
+  /* Clear out existing values */
+  m_uri_scheme.clear();
+  m_uri_host.clear();
+  m_uri_port = 0;
+  m_uri_path.clear();
+
+  /* If already connected, ClusterMgr will send new ProcessInfo reports.
+     Otherwise save a copy of values until connected.
+  */
+  if(m_transporter_facade->theClusterMgr->getNoOfConnectedNodes())
+  {
+    m_transporter_facade->theClusterMgr->setProcessInfoUri(scheme, host,
+                                                           port, path);
+  }
+  else
+  {
+    m_uri_scheme.assign(scheme);
+    m_uri_host.assign(host);
+    m_uri_port = port;
+    m_uri_path.assign(path);
+  }
+
+  return 0;
+}
+
 int
 Ndb_cluster_connection_impl::init_nodes_vector(Uint32 nodeid,
 					       const ndb_mgm_configuration 
@@ -637,52 +998,55 @@ Ndb_cluster_connection_impl::init_nodes_vector(Uint32 nodeid,
       break;
     }
     }
-    if (m_all_nodes.push_back(Node(group,remoteNodeId)))
+    m_db_nodes.set(remoteNodeId);
+    if (m_nodes_proximity.push_back(Node(group,remoteNodeId)))
     {
       DBUG_RETURN(-1);
     }
     DBUG_PRINT("info",("saved %d %d", group,remoteNodeId));
-    for (int i= m_all_nodes.size()-2;
-	 i >= 0 && m_all_nodes[i].group > m_all_nodes[i+1].group;
+    for (int i= m_nodes_proximity.size()-2;
+	 i >= 0 &&
+           m_nodes_proximity[i].adjusted_group >
+             m_nodes_proximity[i+1].adjusted_group;
 	 i--)
     {
-      Node tmp= m_all_nodes[i];
-      m_all_nodes[i]= m_all_nodes[i+1];
-      m_all_nodes[i+1]= tmp;
+      Node tmp= m_nodes_proximity[i];
+      m_nodes_proximity[i]= m_nodes_proximity[i+1];
+      m_nodes_proximity[i+1]= tmp;
     }
   }
 
   int i;
-  Uint32 cur_group, i_group= 0;
-  cur_group= ~0;
-  for (i= (int)m_all_nodes.size()-1; i >= 0; i--)
+  Int32 cur_group= INVALID_PROXIMITY_GROUP;
+  Uint32 group_idx= 0;
+  for (i= (int)m_nodes_proximity.size()-1; i >= 0; i--)
   {
-    if (m_all_nodes[i].group != cur_group)
+    if (m_nodes_proximity[i].adjusted_group != cur_group)
     {
-      cur_group= m_all_nodes[i].group;
-      i_group= i+1;
+      cur_group= m_nodes_proximity[i].adjusted_group;
+      group_idx= i+1;
     }
-    m_all_nodes[i].next_group= i_group;
+    m_nodes_proximity[i].next_group_idx= group_idx;
   }
-  cur_group= ~0;
-  for (i= 0; i < (int)m_all_nodes.size(); i++)
+  cur_group= INVALID_PROXIMITY_GROUP;
+  for (i= 0; i < (int)m_nodes_proximity.size(); i++)
   {
-    if (m_all_nodes[i].group != cur_group)
+    if (m_nodes_proximity[i].adjusted_group != cur_group)
     {
-      cur_group= m_all_nodes[i].group;
-      i_group= i;
+      cur_group= m_nodes_proximity[i].adjusted_group;
+      group_idx= i;
     }
-    m_all_nodes[i].this_group= i_group;
+    m_nodes_proximity[i].this_group_idx= group_idx;
   }
 #if 0
-  for (i= 0; i < (int)m_all_nodes.size(); i++)
+  for (i= 0; i < (int)m_nodes_proximity.size(); i++)
   {
     fprintf(stderr, "[%d] %d %d %d %d\n",
 	   i,
-	   m_all_nodes[i].id,
-	   m_all_nodes[i].group,
-	   m_all_nodes[i].this_group,
-	   m_all_nodes[i].next_group);
+	   m_nodes_proximity[i].id,
+	   m_nodes_proximity[i].adjusted_group,
+	   m_nodes_proximity[i].this_group_idx,
+	   m_nodes_proximity[i].next_group_idx);
   }
 
   do_test();
@@ -693,14 +1057,61 @@ Ndb_cluster_connection_impl::init_nodes_vector(Uint32 nodeid,
 Uint32
 Ndb_cluster_connection_impl::get_db_nodes(Uint8 arr[MAX_NDB_NODES]) const
 {
-  Uint32 cnt = (Uint32)m_all_nodes.size();
-  assert(cnt < MAX_NDB_NODES);
-  const Node *nodes = m_all_nodes.getBase();
-  for (Uint32 i = 0; i<cnt; i++)
-    arr[i] = (Uint8)nodes[i].id;
+  require(m_db_nodes.count() < MAX_NDB_NODES);
+  Uint32 cnt = 0;
+  for (Uint32 node_id = m_db_nodes.find_first();
+       node_id != NdbNodeBitmask::NotFound;
+       node_id = m_db_nodes.find_next(node_id + 1))
+  {
+    arr[cnt] = node_id;
+    cnt++;
+  }
   return cnt;
 }
 
+Uint32
+Ndb_cluster_connection_impl::get_unconnected_nodes() const
+{
+  TransporterFacade *tp = m_transporter_facade;
+
+  NdbNodeBitmask connected; // All nodes connected
+  NdbNodeBitmask started;   // All started nodes known by connected db nodes
+
+  tp->lock_poll_mutex();
+  for (Uint32 node_id = m_db_nodes.find_first();
+       node_id != NdbNodeBitmask::NotFound;
+       node_id = m_db_nodes.find_next(node_id + 1))
+  {
+    const trp_node& node = tp->theClusterMgr->getNodeInfo(node_id);
+    if (!node.m_alive)
+    {
+      continue;
+    }
+    connected.set(node_id);
+    NdbNodeBitmask nodes;
+    // Truncate NodeBitmask to NdbNodeBitmask, data nodes are in lower bits
+    nodes.assign(nodes.Size, node.m_state.m_connected_nodes.rep.data);
+    started.bitOR(nodes);
+  }
+  tp->unlock_poll_mutex();
+
+  if (connected.count() == 0)
+  {
+    /**
+     * No db nodes connected, means all unconnected.
+     */
+    assert(m_db_nodes.count() ==
+             m_nodes_proximity.size());
+    return m_nodes_proximity.size();
+  }
+
+  /**
+   * Return count of started but not connected db nodes
+   */
+  started.bitAND(m_db_nodes);
+  return started.bitANDC(connected).count();
+}
+
 int
 Ndb_cluster_connection_impl::configure(Uint32 nodeId,
                                        const ndb_mgm_configuration &config)
@@ -737,6 +1148,13 @@ Ndb_cluster_connection_impl::configure(Uint32 nodeId,
     {
       m_config.m_default_hashmap_size = default_hashmap_size;
     }
+
+    Uint32 verbose= 0;
+    if (!iter.get(CFG_API_VERBOSE, &verbose))
+    {
+      m_config.m_verbose = verbose;
+    }
+
     // If DefaultHashmapSize is not set or zero, use the minimum
     // value set (not zero) for any other node, since this size
     // should be supported by the other nodes.  Also this allows
@@ -778,6 +1196,13 @@ Ndb_cluster_connection_impl::configure(Uint32 nodeId,
       m_config.m_waitfor_timeout = timeout;
     }
   }
+
+  // System name
+  ndb_mgm_configuration_iterator s_iter(config, CFG_SECTION_SYSTEM);
+  const char * tmp_system_name;
+  s_iter.get(CFG_SYS_NAME, & tmp_system_name);
+  m_system_name.assign(tmp_system_name);
+
   DBUG_RETURN(init_nodes_vector(nodeId, config));
 }
 
@@ -833,11 +1258,28 @@ Ndb_cluster_connection_impl::do_test()
   delete [] nodes;
 }
 
+void Ndb_cluster_connection::set_data_node_neighbour(Uint32 node)
+{
+  m_impl.set_data_node_neighbour(node);
+}
+
 void Ndb_cluster_connection::set_name(const char *name)
 {
   m_impl.set_name(name);
 }
 
+int Ndb_cluster_connection::set_service_uri(const char * scheme,
+                                            const char * host, int port,
+                                            const char * path)
+{
+  return m_impl.set_service_uri(scheme, host, port, path);
+}
+
+const char * Ndb_cluster_connection::get_system_name() const
+{
+  return m_impl.m_system_name.c_str();
+}
+
 int Ndb_cluster_connection_impl::connect(int no_retries,
                                          int retry_delay_in_seconds,
                                          int verbose)
@@ -903,7 +1345,10 @@ int Ndb_cluster_connection_impl::connect(int no_retries,
       ndb_mgm_destroy_configuration(props);
       DBUG_RETURN(-1);
     }
-
+    m_transporter_facade->theClusterMgr->setProcessInfoUri(m_uri_scheme.c_str(),
+                                                           m_uri_host.c_str(),
+                                                           m_uri_port,
+                                                           m_uri_path.c_str());
     ndb_mgm_destroy_configuration(props);
     m_transporter_facade->connected();
     m_latest_error = 0;
@@ -938,16 +1383,15 @@ void Ndb_cluster_connection_impl::connect_thread()
   DBUG_ENTER("Ndb_cluster_connection_impl::connect_thread");
   int r;
   do {
+    // Wait before making a new connect attempt
     NdbSleep_SecSleep(1);
+
     if ((r = connect(0,0,0)) == 0)
       break;
     if (r == -1) {
       printf("Ndb_cluster_connection::connect_thread error\n");
-      DBUG_ASSERT(false);
+      assert(false);
       m_run_connect_thread= 0;
-    } else {
-      // Wait before making a new connect attempt
-      NdbSleep_SecSleep(1);
     }
   } while (m_run_connect_thread);
   if (m_connect_callback)
@@ -1076,6 +1520,119 @@ Ndb_cluster_connection::release_ndb_wait_group(NdbWaitGroup *group)
   }
 }
 
+Uint32
+Ndb_cluster_connection_impl::select_node(const Uint16 * nodes,
+                                         Uint32 cnt)
+{
+  if (cnt == 1)
+  {
+    return nodes[0];
+  }
+  else if (cnt == 0)
+  {
+    return 0;
+  }
+
+  NdbNodeBitmask checked;
+  Node *nodes_arr = m_nodes_proximity.getBase();
+  const Uint32 nodes_arr_cnt = m_nodes_proximity.size();
+
+  Uint32 best_node = nodes[0];
+  Uint32 best_idx;
+  Uint32 best_usage;
+  Int32 best_score = MAX_PROXIMITY_GROUP; // Lower is better
+
+  if (!m_impl.m_optimized_node_selection)
+  {
+    /**
+     * optimized_node_selection is off.  Use round robin.
+     * Uses hint_count in m_nodes_proximity but not the group value.
+     */
+    for (Uint32 j = 0; j < cnt; j++)
+    {
+      Uint32 candidate_node = nodes[j];
+      if (checked.get(candidate_node))
+        continue;
+
+      checked.set(candidate_node);
+
+      for (Uint32 i = 0; i < nodes_arr_cnt; i++)
+      {
+        if (nodes_arr[i].id == j)
+        {
+          Uint32 usage = nodes_arr[i].hint_count;
+          if (best_score == MAX_PROXIMITY_GROUP)
+          {
+            best_idx = i;
+            best_node = candidate_node;
+            best_score = 0;
+            best_usage = usage;
+          }
+          else if (best_usage - usage < HINT_COUNT_HALF)
+          {
+            best_idx = i;
+            best_node = candidate_node;
+            best_usage = usage;
+          }
+          break;
+        }
+      }
+    }
+  }
+  else
+  {
+    /**
+     * optimized_node_selection is on.  Use proximity.
+     */
+    for (Uint32 j = 0; j < cnt; j++)
+    {
+      Uint32 candidate_node = nodes[j];
+      if (checked.get(candidate_node))
+        continue;
+
+      checked.set(candidate_node);
+
+      for (Uint32 i = 0; i < nodes_arr_cnt; i++)
+      {
+        if (nodes_arr[i].adjusted_group > best_score)
+        {
+          // We already got a better match
+          break;
+        }
+        if (nodes_arr[i].id == candidate_node)
+        {
+          if (nodes_arr[i].adjusted_group < best_score)
+          {
+            best_idx = i;
+            best_node = candidate_node;
+            best_score = nodes_arr[i].adjusted_group;
+            best_usage = nodes_arr[i].hint_count;
+          }
+          else if (nodes_arr[i].adjusted_group == best_score)
+          {
+            Uint32 usage = nodes_arr[i].hint_count;
+            if (best_usage - usage < HINT_COUNT_HALF)
+            {
+              /**
+               * hint_count may wrap, for this calculation it is assummed that
+               * the two counts should be near each other, and so if the
+               * difference is small above, best_usage is greater than usage.
+               */
+              best_idx = i;
+              best_node = candidate_node;
+              best_usage = usage;
+            }
+          }
+          break;
+        }
+      }
+    }
+  }
+  nodes_arr[best_idx].hint_count =
+    (nodes_arr[best_idx].hint_count + 1) & HINT_COUNT_MASK;
+  return best_node;
+}
+
 template class Vector<Ndb_cluster_connection_impl::Node>;
 
 int
@@ -1112,18 +1669,20 @@ Ndb_cluster_connection::wait_until_ready(const int * nodes, int cnt,
   {
     dead.clear();
     alive.clear();
-    tp->lock_mutex();
-    for(unsigned i= 0; i < no_db_nodes(); i++)
+    tp->lock_poll_mutex();
+    for (Uint32 node_id = m_impl.m_db_nodes.find_first();
+         node_id != NdbNodeBitmask::NotFound;
+         node_id = m_impl.m_db_nodes.find_next(node_id + 1))
     {
       //************************************************
       // If any node is answering, ndb is answering
       //************************************************
-      if (tp->get_node_alive(m_impl.m_all_nodes[i].id) != 0)
-        alive.set(m_impl.m_all_nodes[i].id);
+      if (tp->get_node_alive(node_id) != 0)
+        alive.set(node_id);
       else
-        dead.set(m_impl.m_all_nodes[i].id);
+        dead.set(node_id);
     }
-    tp->unlock_mutex();
+    tp->unlock_poll_mutex();
 
     if (alive.contains(mask))
     {
