@@ -1,14 +1,21 @@
 /*
-  Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
+  Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
   This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License as published by
-  the Free Software Foundation; version 2 of the License.
+  it under the terms of the GNU General Public License, version 2.0,
+  as published by the Free Software Foundation.
+
+  This program is also distributed with certain software (including
+  but not limited to OpenSSL) that is licensed under separate terms,
+  as designated in a particular file or component or in included license
+  documentation.  The authors of MySQL hereby grant you an additional
+  permission to link the program and your derivative works with the
+  separately licensed software that they have included with MySQL.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
+  GNU General Public License, version 2.0, for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
@@ -21,11 +28,12 @@
 #include <NdbTick.h>
 #include "../../src/ndbapi/NdbQueryOperation.hpp"
 
-HugoQueries::HugoQueries(const NdbQueryDef & query)
+HugoQueries::HugoQueries(const NdbQueryDef & query, int retryMax)
+ : m_query_def(&query),
+   m_ops(query.getNoOfOperations()),
+   m_retryMax(retryMax),
+   m_error()
 {
-  m_retryMax = 100;
-  m_query_def = &query;
-
   for (Uint32 i = 0; i<query.getNoOfOperations(); i++)
   {
     struct Op op;
@@ -134,10 +142,10 @@ HugoQueries::getValueForQueryOp(NdbQueryOperation* pOp, NDBT_ResultRow * pRow)
 
 int
 HugoQueries::runLookupQuery(Ndb* pNdb,
-                            int records,
+                            int queries,
                             int batch)
 {
-  int r = 0;
+  int q = 0;
   int retryAttempt = 0;
 
   m_rows_found.clear();
@@ -152,10 +160,10 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
 
   allocRows(batch);
 
-  while (r < records)
+  while (q < queries)
   {
-    if(r + batch > records)
-      batch = records - r;
+    if (q + batch > queries)
+      batch = queries - q;
 
     if (retryAttempt >= m_retryMax)
     {
@@ -163,11 +171,16 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
              << " times, failing!" << endl;
       return NDBT_FAILED;
     }
+    if (retryAttempt > 0)
+    {
+      NdbSleep_MilliSleep(50);
+    }
 
     Vector<Uint32> batch_rows_found;
     batch_rows_found.fill(m_query_def->getNoOfOperations() - 1, zero);
     Vector<NdbQuery*> queries;
 
+    clearNdbError();
     NdbTransaction * pTrans = pNdb->startTransaction();
     if (pTrans == NULL)
     {
@@ -175,11 +188,12 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
 
       if (err.status == NdbError::TemporaryError){
         NDB_ERR(err);
-        NdbSleep_MilliSleep(50);
+        setNdbError(err);
         retryAttempt++;
         continue;
       }
       NDB_ERR(err);
+      setNdbError(err);
       return NDBT_FAILED;
     }
 
@@ -187,13 +201,14 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
     {
       char buf[NDB_MAX_TUPLE_SIZE];
       NdbQueryParamValue params[NDB_MAX_NO_OF_ATTRIBUTES_IN_KEY];
-      equalForParameters(buf, m_ops[0], params, b + r);
+      equalForParameters(buf, m_ops[0], params, b + q);
 
       NdbQuery * query = pTrans->createQuery(m_query_def, params);
       if (query == 0)
       {
         const NdbError err = pTrans->getNdbError();
         NDB_ERR(err);
+        setNdbError(err);
         return NDBT_FAILED;
       }
 
@@ -210,9 +225,9 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
     {
       const NdbError err = pTrans->getNdbError();
       NDB_ERR(err);
+      setNdbError(err);
       if (err.status == NdbError::TemporaryError){
         pTrans->close();
-        NdbSleep_MilliSleep(50);
         retryAttempt++;
         continue;
       }
@@ -231,6 +246,7 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
       if (err.code)
       {
         NDB_ERR(err);
+        setNdbError(err);
         ndbout_c("API INCONSISTENCY: NdbTransaction returned NdbError even if ::execute() succeeded");
         pTrans->close();
         return NDBT_FAILED;
@@ -253,6 +269,7 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
       if (err.code)
       {
         NDB_ERR(err);
+        setNdbError(err);
         if (err.status == NdbError::TemporaryError){
           pTrans->close();
           retry = true;
@@ -283,6 +300,7 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
       {
         const NdbError& err = query->getNdbError();
         NDB_ERR(err);
+        setNdbError(err);
         if (err.status == NdbError::TemporaryError){
           pTrans->close();
           retry = true;
@@ -294,13 +312,12 @@ HugoQueries::runLookupQuery(Ndb* pNdb,
     }
     if (retry)
     {
-      NdbSleep_MilliSleep(50);
       retryAttempt++;
       continue;
     }
 
     pTrans->close();
-    r += batch;
+    q += batch;
 
     for (unsigned i = 0; i<batch_rows_found.size(); i++)
       m_rows_found[i] += batch_rows_found[i];
@@ -321,17 +338,22 @@ HugoQueries::runScanQuery(Ndb * pNdb,
 
   while (retryAttempt < m_retryMax)
   {
+    if (retryAttempt > 0)
+    {
+      NdbSleep_MilliSleep(50);
+    }
     m_rows_found.clear();
     Uint32 zero = 0;
     m_rows_found.fill(m_query_def->getNoOfOperations() - 1, zero);
 
+    clearNdbError();
     NdbTransaction * pTrans = pNdb->startTransaction();
     if (pTrans == NULL)
     {
       const NdbError err = pNdb->getNdbError();
       NDB_ERR(err);
+      setNdbError(err);
       if (err.status == NdbError::TemporaryError){
-        NdbSleep_MilliSleep(50);
         retryAttempt++;
         continue;
       }
@@ -348,6 +370,7 @@ HugoQueries::runScanQuery(Ndb * pNdb,
     {
       const NdbError err = pTrans->getNdbError();
       NDB_ERR(err);
+      setNdbError(err);
       return NDBT_FAILED;
     }
 
@@ -362,9 +385,9 @@ HugoQueries::runScanQuery(Ndb * pNdb,
     {
       const NdbError err = pTrans->getNdbError();
       NDB_ERR(err);
+      setNdbError(err);
       if (err.status == NdbError::TemporaryError){
         pTrans->close();
-        NdbSleep_MilliSleep(50);
         retryAttempt++;
         continue;
       }
@@ -383,6 +406,7 @@ HugoQueries::runScanQuery(Ndb * pNdb,
       if (err.code)
       {
         NDB_ERR(err);
+        setNdbError(err);
         ndbout_c("API INCONSISTENCY: NdbTransaction returned NdbError even if ::execute() succeeded");
         pTrans->close();
         return NDBT_FAILED;
@@ -399,9 +423,9 @@ HugoQueries::runScanQuery(Ndb * pNdb,
       if (err.code)
       {
         NDB_ERR(err);
+        setNdbError(err);
         if (err.status == NdbError::TemporaryError){
           pTrans->close();
-          NdbSleep_MilliSleep(50);
           retryAttempt++;
           continue;
         }
@@ -429,7 +453,7 @@ HugoQueries::runScanQuery(Ndb * pNdb,
         query->close();
         pTrans->close();
         m_rows_found.clear();
-      return NDBT_OK;
+        return NDBT_OK;
       }
 
       for (unsigned o = 0; o<m_ops.size(); o++)
@@ -453,9 +477,9 @@ HugoQueries::runScanQuery(Ndb * pNdb,
     if (res == NdbQuery::NextResult_error)
     {
       NDB_ERR(err);
+      setNdbError(err);
       if (err.status == NdbError::TemporaryError)
       {
-        NdbSleep_MilliSleep(50);
         retryAttempt++;
         continue;
       }
@@ -469,8 +493,30 @@ HugoQueries::runScanQuery(Ndb * pNdb,
     break;
   }
 
+  if (m_error.code != 0)  //Still failures after retries
+    return NDBT_FAILED;
+
   return NDBT_OK;
 }
 
+void
+HugoQueries::clearNdbError()
+{
+  m_error.code = 0;
+}
+
+void
+HugoQueries::setNdbError(const NdbError& error)
+{
+  assert(error.code != 0);
+  m_error = error;
+}
+
+const NdbError& 
+HugoQueries::getNdbError() const
+{
+  return m_error;
+}
+
 template class Vector<HugoQueries::Op>;
 template class Vector<NdbQuery*>;
